@@ -24,8 +24,24 @@ new_counters = function(type, rownames) {
   )
 }
 
-# set some internal knitr options
-set_opts_knit = function(config) {
+# set common format config
+common_format_config = function(
+  config, format, file_scope = getOption('bookdown.render.file_scope', FALSE)
+) {
+
+  # provide file_scope if requested
+  if (file_scope) config$file_scope = md_chapter_splitter
+
+  # prepend the custom-environment filter
+  config$pandoc$lua_filters = c(
+    lua_filter("custom-environment.lua"), config$pandoc$lua_filters
+  )
+  # and add bookdown metadata file for the filter to work
+  config$pandoc$args = c(bookdown_yml_arg(), config$pandoc$args)
+
+  # set output format
+  config$bookdown_output_format = format
+
   # use labels of the form (\#label) in knitr
   config$knitr$opts_knit$bookdown.internal.label = TRUE
   # when the output is LaTeX, force LaTeX tables instead of default Pandoc tables
@@ -59,19 +75,30 @@ book_filename = function(config = load_config(), fallback = TRUE) {
 }
 
 source_files = function(format = NULL, config = load_config(), all = FALSE) {
-  # a list of Rmd chapters
   subdir = config[['rmd_subdir']]; subdir_yes = isTRUE(subdir) || is.character(subdir)
+  # a list of Rmd chapters
   files = list.files('.', '[.]Rmd$', ignore.case = TRUE)
-  files = c(files, list.files(
+  # content in subdir if asked
+  subdir_files = unlist(mapply(
+    list.files,
     if (is.character(subdir)) subdir else '.', '[.]Rmd$', ignore.case = TRUE,
-    recursive = subdir_yes, full.names = subdir_yes
+    recursive = subdir_yes, full.names = is.character(subdir), USE.NAMES = FALSE
   ))
+  subdir_files = setdiff(subdir_files, files)
+  files = c(files, subdir_files)
+  # if rmd_files is provided, use those files in addition to those under rmd_subdir
   if (length(files2 <- config[['rmd_files']]) > 0) {
+    # users should specify 'docx' as the output format name for Word, but let's
+    # make 'word' an alias of 'docx' to avoid further confusion:
+    # https://stackoverflow.com/q/63678601/559676
+    if ('word' %in% names(files2) && identical(format, 'docx')) format = 'word'
     if (is.list(files2)) files2 = if (all) unlist(files2) else files2[[format]]
-    files = if (subdir_yes) c(files2, files) else files2
-  } else {
-    files = files[grep('^[^_]', basename(files))]  # exclude those start with _
+    # add those files to subdir content if any
+    files = if (subdir_yes) c(files2, subdir_files) else files2
   }
+  # exclude files that start with _, and the merged file
+  files = files[grep('^[^_]', basename(files))]
+  files = setdiff(files, with_ext(book_filename(config), c('.md', '.Rmd')))
   files = unique(gsub('^[.]/', '', files))
   index = 'index' == with_ext(files, '')
   # if there is a index.Rmd, put it in the beginning
@@ -102,8 +129,9 @@ mark_dirs = function(x) {
   x
 }
 
+# TODO: use xfun::del_empty_dir()
 clean_empty_dir = function(dir) {
-  if (!dir_exists(dir)) return()
+  if (is.null(dir) || !dir_exists(dir)) return()
   files = list.files(dir, all.files = TRUE, recursive = TRUE)
   if (length(files) == 0) unlink(dir, recursive = TRUE)
 }
@@ -127,6 +155,25 @@ merge_chapters = function(files, to, before = NULL, after = NULL, orig = files)
   Sys.chmod(to, '644')
 }
 
+# split a markdown file into a set of chapters
+md_chapter_splitter = function(file) {
+  x = read_utf8(file)
+
+  # get positions of the chapter delimiters (r_chap_pattern defined in html.R)
+  if (length(pos <- grep(r_chap_pattern, x)) <= 1) return()
+  pos = c(0, pos)
+
+  # get the filenames
+  names = gsub(r_chap_pattern, '\\1', x[pos])
+
+  # extract the chapters and pair them w/ the names
+  lapply(seq_along(names), function(i) {
+    i1 = pos[i] + 1
+    i2 = pos[i + 1]
+    list(name = names[i], content = x[i1:i2])
+  })
+}
+
 match_dashes = function(x) grep('^---\\s*$', x)
 
 create_placeholder = function(x) {
@@ -163,12 +210,25 @@ insert_code_chunk = function(x, before, after) {
 }
 
 insert_chapter_script = function(config, where = 'before') {
-  script = config[[sprintf('%s_chapter_script', where)]]
+  script = get_chapter_script(config, where)
   if (is.character(script)) {
-    c('```{r include=FALSE, cache=FALSE}', unlist(lapply(script, read_utf8)), '```')
+    c('```{r include=FALSE, cache=FALSE}', script, '```')
   }
 }
 
+get_chapter_script = function(config, where) {
+  script = config[[sprintf('%s_chapter_script', where)]]
+  unlist(lapply(script, read_utf8))
+}
+
+merge_chapter_script = function(config, where) {
+  if (!is.character(script <- get_chapter_script(config, where)) || length(script) == 0)
+    return('')
+  f = tempfile(fileext = '.R')
+  write_utf8(script, f)
+  f
+}
+
 check_special_chars = function(filename) {
   reg = rmarkdown:::.shell_chars_regex
   for (i in grep(reg, filename)) warning(
@@ -186,6 +246,11 @@ Rscript_render = function(file, ...) {
   if (Rscript(args) != 0) stop('Failed to compile ', file)
 }
 
+source_utf8 = function(file) {
+  if (file == '') return()
+  eval(xfun::parse_only(read_utf8(file)), envir = globalenv())
+}
+
 clean_meta = function(meta_file, files) {
   meta = readRDS(meta_file)
   for (i in setdiff(names(meta), files)) meta[[i]] = NULL
@@ -208,6 +273,7 @@ strip_search_text = function(x) {
   x = gsub('<script[^>]*>(.*?)</script>', '', x)
   x = gsub('<div id="refs" class="references">.*', '', x)
   x = strip_html(x)
+  x = gsub('[[:space:]]', ' ', x)
   x
 }
 
@@ -228,12 +294,29 @@ local_resources = function(x) {
   grep('^(f|ht)tps?://.+', x, value = TRUE, invert = TRUE)
 }
 
-#' Continously preview the HTML output of a book using the \pkg{servr} package
+# write out reference keys to _book/reference-keys.txt (for the RStudio visual
+# editor to autocomplete \@ref())
+write_ref_keys = function(x) {
+  # this only works for books rendered with bookdown::render_book() (and not for
+  # rmarkdown::render())
+  if (is.null(preview <- opts$get('preview'))) return()
+  # collect reference keys from parse_fig_labels() and parse_section_labels()
+  if (is.null(d <- opts$get('output_dir'))) return()
+  p = ref_keys_path(d)
+  if (file.exists(p)) x = unique(c(xfun::read_utf8(p), x))
+  xfun::write_utf8(x, p)
+}
+
+ref_keys_path = function(d = opts$get('output_dir')) {
+  file.path(d, 'reference-keys.txt')
+}
+
+#' Continuously preview the HTML output of a book using the \pkg{servr} package
 #'
 #' When any files are modified or added to the book directory, the book will be
 #' automatically recompiled, and the current HTML page in the browser will be
-#' refreshed. This function is based on \code{servr::\link[servr]{httw}()} to
-#' continuously watch a directory.
+#' refreshed. This function is based on \code{servr::\link[servr:httd]{httw}()}
+#' to continuously watch a directory.
 #'
 #' For \code{in_session = TRUE}, you will have access to all objects created in
 #' the book in the current R session: if you use a daemonized server (via the
@@ -262,8 +345,8 @@ local_resources = function(x) {
 #'   the book directory.
 #' @param quiet Whether to suppress output (e.g., the knitting progress) in the
 #'   console.
-#' @param ... Other arguments passed to \code{servr::\link[servr]{httw}()} (not
-#'   including the \code{handler} argument, which has been set internally).
+#' @param ... Other arguments passed to \code{servr::\link[servr:httd]{httw}()}
+#'   (not including the \code{handler} argument, which has been set internally).
 #' @export
 serve_book = function(
   dir = '.', output_dir = '_book', preview = TRUE, in_session = TRUE, quiet = FALSE, ...
@@ -311,7 +394,7 @@ serve_book = function(
 first_html_format = function() {
   fallback = 'bookdown::gitbook'
   if (!file.exists('index.Rmd')) return(fallback)
-  formats = rmarkdown::all_output_formats('index.Rmd', 'UTF-8')
+  formats = rmarkdown::all_output_formats('index.Rmd')
   formats = grep('gitbook|html', formats, value = TRUE)
   if (length(formats) == 0) fallback else formats[1]
 }
@@ -483,3 +566,68 @@ strip_latex_body = function(x, alt = '\nThe content was intentionally removed.\n
   }
   c(x1, x2[sort(i)], '\\end{document}')
 }
+
+# bookdown Lua filters paths
+lua_filter = function (filters = NULL) {
+  rmarkdown::pkg_file_lua(filters, package = 'bookdown')
+}
+
+# pass _bookdown.yml to Pandoc's Lua filters
+bookdown_yml_arg = function(config = load_config(), path = tempfile()) {
+  # this is supported for Pandoc >= 2.0 only
+  if (!pandoc2.0() || length(config) == 0) return()
+  yaml::write_yaml(list(bookdown = config), path)
+  c("--metadata-file", rmarkdown::pandoc_path_arg(path))
+}
+
+#' Convert the syntax of theorem and proof environments from code blocks to
+#' fenced Divs
+#'
+#' This function converts the syntax \samp{```{theorem, label, ...}} to
+#' \samp{::: {.theorem #label ...}} (Pandoc's fenced Div) for theorem
+#' environments.
+#' @param input Path to an Rmd file that contains theorem environments written
+#'   in the syntax of code blocks.
+#' @param text A character vector of the Rmd source. When \code{text} is
+#'   provided, the \code{input} argument will be ignored.
+#' @param output The output file to write the converted input content. You can
+#'   specify \code{output} to be identical to \code{input}, which means the
+#'   input file will be overwritten. If you want to overwrite the input file,
+#'   you are strongly recommended to put the file under version control or make
+#'   a backup copy in advance.
+#' @references Learn more about
+#'   \href{https://bookdown.org/yihui/bookdown/markdown-extensions-by-bookdown.html#theorems}{theorems
+#'    and proofs} and
+#'   \href{https://bookdown.org/yihui/rmarkdown-cookbook/custom-blocks.html}{custom
+#'    blocks} in the \pkg{bookdown} book.
+#' @return If \code{output = NULL}, the converted text is returned, otherwise
+#'   the text is written to the output file.
+#' @export
+fence_theorems = function(input, text = xfun::read_utf8(input), output = NULL) {
+  # identify blocks
+  md_pattern = knitr::all_patterns$md
+  block_start = grep(md_pattern$chunk.begin, text)
+  # extract params
+  params = gsub(md_pattern$chunk.begin, "\\1", text[block_start])
+  # find block with custom environment engine
+  reg = sprintf("^(%s).*", paste(all_math_env, collapse = "|"))
+  to_convert = grepl(reg, params)
+  # only modify those blocks
+  params = params[to_convert]
+  block_start = block_start[to_convert]
+  block_end = grep(md_pattern$chunk.end, text)
+  block_end = vapply(block_start, function(x) block_end[block_end > x][1], integer(1))
+  # add a . to engine name
+  params = sprintf(".%s", params)
+  # change label to id
+  params = gsub(",\\s*([-/[:alnum:]]+)(,|\\s*$)", " #\\1", params)
+  params = gsub("label\\s*=\\s*\"([-/[:alnum:]]+)\"", "#\\1", params)
+  # clean , and space
+  params = gsub(",\\s*", " ", params)
+  params = gsub("\\s*=\\s*", "=", params)
+  # modify the blocks
+  text[block_start] = sprintf("::: {%s}", params)
+  text[block_end] = ":::"
+  # return the text or write to output file
+  if (is.null(output)) xfun::raw_string(text) else xfun::write_utf8(text, input)
+}
